/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */

/// <reference types="cypress" />

import UserRepresentation from '@keycloak/keycloak-admin-client/lib/defs/userRepresentation';
import { getCypressEnv } from './cypress-helper';

enum HttpMethod {
  POST = 'POST',
  PUT = 'PUT',
  GET = 'GET',
}

const DATA_TEST_ID: string = 'data-test-id';
const DATA_TEST_CLASS: string = 'data-test-class';

const ACCES_TOKEN: string = 'access_token';

enum Header {
  CONTENT_TYPE = 'Content-Type',
  AUTHORIZATION = 'Authorization',
}

enum CypressEnv {
  KEYCLOAK_CLIENT = 'keycloakClient',
  KEYCLOAK_REALM = 'keycloakRealm',
  KEYCLOAK_URL = 'keycloakUrl',
  SEARCH = 'search',
  SMOCKER = 'smocker',
}

Cypress.Commands.add('getTestElement', (selector, ...args) => {
  return cy.get(`[${DATA_TEST_ID}~="${selector}"]`, ...args);
});

Cypress.Commands.add('getChildTestElementFromListElement', (parentSelector, index, childSelector, ...args) => {
  return cy
    .get(`[${DATA_TEST_ID}~="${parentSelector}"]`, ...args)
    .eq(index)
    .find(`[${DATA_TEST_ID}~="${childSelector}"]`);
});

Cypress.Commands.add('getTestElementWithClass', (selector, ...args) => {
  console.log(
    'Achtung: Potentiell nicht eindeutiges Ergebnis, weil eine data-test-class mit cy.get() von der DOM-Root aus gesucht wird.',
  );
  return cy.get(`[${DATA_TEST_CLASS}="${selector}"]`, ...args);
});

Cypress.Commands.add('getTestElementWithOid', (oid, ...args) => {
  return cy.getTestElement(oid, ...args);
});

Cypress.Commands.add('findTestElementWithClass', { prevSubject: true }, (subject: any, selector) => {
  return subject.find(`[${DATA_TEST_CLASS}="${selector}"]`);
});

Cypress.Commands.add('findElement', { prevSubject: true }, (subject: any, selector: string) => {
  return subject.find(selector);
});

Cypress.Commands.add('getUserInfo', () => {
  return cy.request({
    method: HttpMethod.GET,
    url: `${getKeycloakBaseRealmUrl()}/userinfo`,
    headers: { ...buildBearerTokenAuthorizationHeader() },
  });
});

Cypress.Commands.add('findUser', (username: string) =>
  cy
    .request<UserRepresentation[]>({
      method: HttpMethod.GET,
      url: getUsersApiUrl(),
      headers: { ...buildBearerTokenAuthorizationHeader() },
    })
    .then<UserRepresentation>((response: Cypress.Response<UserRepresentation[]>) => {
      const users: UserRepresentation[] = response.body;
      const found: UserRepresentation | undefined = users.find((u: UserRepresentation) => u.username === username);
      if (found) return found;
      return Promise.reject(`User with username ${username} not found`);
    }),
);

Cypress.Commands.add('verifyEmail', (userId: string) =>
  cy.request<void>({
    method: HttpMethod.PUT,
    url: getUsersApiUrl() + '/' + userId,
    headers: { ...buildBearerTokenAuthorizationHeader() },
    body: { emailVerified: true },
  }),
);

Cypress.Commands.add('resetPassword', (userId: string, newPassword: string) =>
  cy.request<void>({
    method: HttpMethod.PUT,
    url: getUsersApiUrl() + '/' + userId + '/reset-password',
    headers: { ...buildBearerTokenAuthorizationHeader() },
    body: { type: 'password', temporary: false, value: newPassword },
  }),
);

function buildBearerTokenAuthorizationHeader(): object {
  return {
    [Header.AUTHORIZATION]: `Bearer ${getAccessToken()}`,
  };
}

function getAccessToken(): string | null {
  return window.sessionStorage.getItem(ACCES_TOKEN);
}

function getUsersApiUrl(): string {
  return getCypressEnv('keycloakUrl') + '/admin/realms/' + getCypressEnv('keycloakRealm') + '/users';
}

function getKeycloakBaseRealmUrl(): string {
  return `${Cypress.env(CypressEnv.KEYCLOAK_URL)}realms/${Cypress.env(CypressEnv.KEYCLOAK_REALM)}/protocol/openid-connect`;
}

type GetOptions = Partial<Cypress.Loggable & Cypress.Timeoutable & Cypress.Withinable & Cypress.Shadow>;

declare global {
  // eslint-disable-next-line
  namespace Cypress {
    interface Chainable {
      getTestElement(selector: string, ...args: GetOptions[]): Chainable<Element>;

      getChildTestElementFromListElement(
        parentSelector: string,
        index: number,
        childSelector: string,
        ...args: GetOptions[]
      ): Chainable<Element>;

      getTestElementWithClass(selector: string, ...args: GetOptions[]): Chainable<Element>;

      getTestElementWithOid(oid: string, ...args: GetOptions[]): Chainable<Element>;

      findTestElementWithClass(selector: string): Chainable<Element>;

      findElement(selector: string): Chainable<Element>;

      getUserInfo(): Chainable<void>;

      findUser(username: string): Chainable<UserRepresentation>;

      verifyEmail(userId: string): Chainable<Response<void>>;

      resetPassword(userId: string, password: string): Chainable<Response<void>>;
    }
  }
}
