/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
import { AdminOrganisationsEinheit, AdminOrganisationsEinheitService } from '@admin-client/organisations-einheit-shared';
import { ROUTES } from '@admin-client/shared';
import { User, UserService } from '@admin-client/user-shared';
import { KeycloakErrorMessage, KeycloakFieldName, PatchConfig } from '@admin/keycloak-shared';
import { NavigationService } from '@alfa-client/navigation-shared';
import {
  createEmptyStateResource,
  createLoadingStateResource,
  createStateResource,
  StateResource,
} from '@alfa-client/tech-shared';
import { Mock, mock, mockWindowError } from '@alfa-client/test-utils';
import { SnackBarService } from '@alfa-client/ui';
import { fakeAsync, TestBed, tick } from '@angular/core/testing';
import { AbstractControl, FormControl, FormGroup, UntypedFormBuilder, UntypedFormGroup } from '@angular/forms';
import { ActivatedRoute, UrlSegment } from '@angular/router';
import { faker } from '@faker-js/faker/locale/de';
import { expect } from '@jest/globals';
import { createUser } from 'libs/admin/user-shared/test/user';
import { Observable, of } from 'rxjs';
import { createUrlSegment } from '../../../../../navigation-shared/test/navigation-test-factory';
import { singleCold, singleColdCompleted, singleHot } from '../../../../../tech-shared/test/marbles';
import { createKeycloakHttpErrorResponse } from '../../../../keycloak-shared/src/test/keycloak';
import { createAdminOrganisationsEinheit } from '../../../../organisations-einheit-shared/src/test/organisations-einheit';
import { UserFormService } from './user.formservice';
import SpyInstance = jest.SpyInstance;

describe('UserFormService', () => {
  let service: UserFormService;

  let roleGroup: UntypedFormGroup;
  let alfaGroup: UntypedFormGroup;
  let organisationsEinheitenGroup: UntypedFormGroup;
  let administrationGroup: UntypedFormGroup;

  let userService: Mock<UserService>;
  let adminOrganisationsEinheitService: Mock<AdminOrganisationsEinheitService>;
  let navigationService: Mock<NavigationService>;
  let snackBarService: Mock<SnackBarService>;
  let activatedRoute: Mock<ActivatedRoute>;

  const adminOrganisationsEinheit: AdminOrganisationsEinheit = createAdminOrganisationsEinheit();
  const adminOrganisationsEinheitList: StateResource<AdminOrganisationsEinheit[]> = createStateResource([
    adminOrganisationsEinheit,
  ]);

  beforeEach(() => {
    userService = {
      ...mock(UserService),
      refresh: jest.fn(),
      create: jest.fn(),
      save: jest.fn(),
      delete: jest.fn(),
      getUserById: jest.fn(),
    };
    adminOrganisationsEinheitService = {
      ...mock(AdminOrganisationsEinheitService),
      getAll: jest.fn().mockReturnValue(of(adminOrganisationsEinheitList)),
    };
    navigationService = mock(NavigationService);
    snackBarService = mock(SnackBarService);
    activatedRoute = <any>{ ...mock(ActivatedRoute), url: of<UrlSegment[]>([]) };

    TestBed.configureTestingModule({
      providers: [
        UserFormService,
        UntypedFormBuilder,
        { provide: UserService, useValue: userService },
        { provide: AdminOrganisationsEinheitService, useValue: adminOrganisationsEinheitService },
        { provide: NavigationService, useValue: navigationService },
        { provide: SnackBarService, useValue: snackBarService },
        { provide: ActivatedRoute, useValue: activatedRoute },
      ],
    });

    service = TestBed.inject(UserFormService);
    organisationsEinheitenGroup = <UntypedFormGroup>service.form.get(UserFormService.GROUPS);
    roleGroup = <UntypedFormGroup>service.form.get(UserFormService.CLIENT_ROLES);
    alfaGroup = <UntypedFormGroup>roleGroup.get(UserFormService.ALFA_GROUP);
    administrationGroup = <UntypedFormGroup>roleGroup.get(UserFormService.ADMINISTRATION_GROUP);
  });

  it('should create', () => {
    expect(service).toBeTruthy();
  });

  describe('build patch config', () => {
    describe('on matching route', () => {
      it('should contains id and value for patch indication', () => {
        const benutzerUrlSegment: UrlSegment = <UrlSegment>{ ...createUrlSegment(), path: 'benutzer' };
        const idUrlSegment: UrlSegment = <UrlSegment>{ ...createUrlSegment(), path: 'dummyId' };

        const patchConfig: PatchConfig = service._buildPatchConfig([benutzerUrlSegment, idUrlSegment]);

        expect(patchConfig).toEqual({ id: 'dummyId', doPatch: true });
      });
    });

    it('should contains value for patch indication only on non matching route', () => {
      const benutzerUrlSegment: UrlSegment = <UrlSegment>{ ...createUrlSegment(), path: 'benutzer' };
      const idUrlSegment: UrlSegment = <UrlSegment>{ ...createUrlSegment(), path: 'neu' };

      const patchConfig: PatchConfig = service._buildPatchConfig([benutzerUrlSegment, idUrlSegment]);

      expect(patchConfig).toEqual({ doPatch: false });
    });
  });

  describe('load', () => {
    const id: string = 'dummyId';

    const loadedUser: StateResource<User> = createStateResource(createUser());

    beforeEach(() => {
      userService.getUserById.mockReturnValue(singleHot(loadedUser));
    });

    it('should call service to get user by id', () => {
      service._load(id);

      expect(userService.getUserById).toHaveBeenCalledWith(id);
    });

    it('should return loaded user', () => {
      const response: Observable<StateResource<User>> = service._load(id);

      expect(response).toBeObservable(singleCold(loadedUser));
    });
  });

  describe('listenToAlfaGroupChanges', () => {
    it('should call handleAlfaGroupChange on initial change', () => {
      service._handleAlfaGroupChange = jest.fn();

      service.listenToAlfaGroupChanges();

      expect(service._handleAlfaGroupChange).toHaveBeenCalled();
    });

    it('should call handleAlfaGroupChange when value of form element changes', fakeAsync(() => {
      service._handleAlfaGroupChange = jest.fn();

      alfaGroup.get(UserFormService.LOESCHEN).setValue(true);

      tick();

      expect(service._handleAlfaGroupChange).toHaveBeenCalled();
    }));
  });

  describe('initOrganisationsEinheiten', () => {
    beforeEach(() => {
      service._addOrganisationsEinheitenToForm = jest.fn();
    });

    it('should call adminOrganisationsEinheitService getAll', () => {
      service._initOrganisationsEinheiten();

      expect(adminOrganisationsEinheitService.getAll).toHaveBeenCalled();
    });

    it('should return resource', () => {
      const result: Observable<AdminOrganisationsEinheit[]> = service._initOrganisationsEinheiten();

      expect(result).toBeObservable(singleColdCompleted(adminOrganisationsEinheitList.resource));
    });

    it('should call addOrganisationsEinheitenToForm ', fakeAsync(() => {
      service._initOrganisationsEinheiten().subscribe();
      tick();

      expect(service._addOrganisationsEinheitenToForm).toHaveBeenCalled();
    }));

    it('should save ids in map', () => {
      service._initOrganisationsEinheiten().subscribe();

      expect(service._organisationsEinheitToGroupIdMap.get(adminOrganisationsEinheit.name)).toEqual(adminOrganisationsEinheit.id);
    });

    it('should set initOrganisationsEinheiten$', () => {
      expect(service['_initOrganisationsEinheiten$']).toBeDefined();
    });

    it('should not throw any exception on loading state resource', () => {
      adminOrganisationsEinheitService.getAll.mockReturnValue(of(createLoadingStateResource()));
      const errorMock: any = mockWindowError();

      service._initOrganisationsEinheiten();

      expect(errorMock).not.toHaveBeenCalled();
    });
  });

  describe('addOrganisationsEinheitenToForm', () => {
    it('should add organisationsEinheiten to form', () => {
      const organisationsEinheitenGroup: UntypedFormGroup = <UntypedFormGroup>service.form.get(UserFormService.GROUPS);

      service._addOrganisationsEinheitenToForm(adminOrganisationsEinheitList.resource, organisationsEinheitenGroup);

      expect(organisationsEinheitenGroup.value).toEqual({ [adminOrganisationsEinheit.name]: false });
    });
  });

  describe('handleAlfaGroupChange', () => {
    it('should call disableUncheckedCheckboxes if any checkbox is checked', () => {
      service._isAnyChecked = jest.fn().mockReturnValue(true);
      service._disableUncheckedCheckboxes = jest.fn();

      service._handleAlfaGroupChange(alfaGroup);

      expect(service._disableUncheckedCheckboxes).toHaveBeenCalled();
    });

    it('should call enableAllCheckboxes if not any checkbox is checked', () => {
      service._isAnyChecked = jest.fn().mockReturnValue(false);
      service._enableAllCheckboxes = jest.fn();

      service._handleAlfaGroupChange(alfaGroup);

      expect(service._enableAllCheckboxes).toHaveBeenCalled();
    });
  });

  describe('isAnyChecked', () => {
    it('should return false if no checkbox is checked', () => {
      const result = service._isAnyChecked(alfaGroup);

      expect(result).toBe(false);
    });

    it('should return true if any checkbox is checked', () => {
      alfaGroup.get(UserFormService.LOESCHEN).setValue(true);

      const result = service._isAnyChecked(alfaGroup);

      expect(result).toBe(true);
    });
  });

  describe('disableUncheckedCheckboxes', () => {
    it('if control value is false then control should be disabled', () => {
      const control: AbstractControl = alfaGroup.get(UserFormService.LOESCHEN);
      control.setValue(false);

      service._disableUncheckedCheckboxes(alfaGroup);

      expect(control.disabled).toBe(true);
    });

    it('if control value is true then control should NOT be disabled', () => {
      const control: AbstractControl = alfaGroup.get(UserFormService.LOESCHEN);
      control.setValue(true);

      service._disableUncheckedCheckboxes(alfaGroup);

      expect(control.disabled).toBe(false);
    });
  });

  describe('updateCheckboxStates', () => {
    it('if control value is false then control should be disabled', () => {
      const control: AbstractControl = alfaGroup.get(UserFormService.LOESCHEN);
      control.setValue(false);

      service._disableUncheckedCheckboxes(alfaGroup);

      expect(control.disabled).toBe(true);
    });
  });

  describe('enableAllCheckboxes', () => {
    it('if control value is true then control should be enabled', () => {
      const control: AbstractControl = alfaGroup.get(UserFormService.LOESCHEN);
      const enableSpy = jest.spyOn(control, 'enable');

      service._enableAllCheckboxes(alfaGroup);

      expect(enableSpy).toHaveBeenCalled();
    });
  });

  describe('doSubmit', () => {
    const user: User = createUser();

    beforeEach(() => {
      service.isInvalid = jest.fn().mockReturnValue(false);
      userService.create.mockReturnValue(of(createStateResource(user)));
    });

    it('should call createUser', () => {
      service._createUser = jest.fn();

      service._doSubmit().subscribe();

      expect(service._createUser).toHaveBeenCalled();
    });

    it('should call _createOrSave', () => {
      service._createUser = jest.fn().mockReturnValue(user);
      service._createOrSave = jest.fn().mockReturnValue(of(user));

      service._doSubmit().subscribe();

      expect(service._createOrSave).toHaveBeenCalledWith(user);
    });

    it('should call handleOnCreateUserSuccess if not loading', fakeAsync(() => {
      const handleOnCreateUserSuccessSpy: SpyInstance = jest.spyOn(service, 'handleOnCreateUserSuccess');

      service._doSubmit().subscribe();

      expect(handleOnCreateUserSuccessSpy).toHaveBeenCalled();
    }));

    it('should not call handleOnCreateUserSuccess if loading', fakeAsync(() => {
      userService.create.mockReturnValue(of(createEmptyStateResource(true)));
      const handleOnCreateUserSuccessSpy: SpyInstance = jest.spyOn(service, 'handleOnCreateUserSuccess');

      service.submit().subscribe();
      tick();

      expect(handleOnCreateUserSuccessSpy).not.toHaveBeenCalled();
    }));
  });

  describe('createOrSave', () => {
    const user: User = createUser();

    const id: string = faker.string.uuid();

    const patchConfig: PatchConfig = { id, doPatch: true };

    it('should call save if patched', () => {
      service.isPatch = jest.fn().mockReturnValue(true);
      service._patchConfig = patchConfig;
      user.groups = [adminOrganisationsEinheit.name];
      service._organisationsEinheitToGroupIdMap.set(adminOrganisationsEinheit.name, adminOrganisationsEinheit.id);

      service._createOrSave(user);

      expect(userService.save).toHaveBeenCalledWith({ ...user, id, groupIds: [adminOrganisationsEinheit.id] });
    });

    it('should call create if not patched', () => {
      service.isPatch = jest.fn().mockReturnValue(false);

      service._createOrSave(user);

      expect(userService.create).toHaveBeenCalledWith(user);
    });
  });

  describe('handleOnCreateUserSuccess', () => {
    it('should show success message for create', () => {
      service.handleOnCreateUserSuccess();

      expect(snackBarService.showInfo).toHaveBeenCalledWith('Der Benutzer wurde angelegt.');
    });

    it('should show success message for patch', () => {
      service.isPatch = jest.fn().mockReturnValue(true);

      service.handleOnCreateUserSuccess();

      expect(snackBarService.showInfo).toHaveBeenCalledWith('Der Benutzer wurde gespeichert.');
    });

    it('should navigate back to user list', () => {
      service.handleOnCreateUserSuccess();

      expect(navigationService.navigate).toHaveBeenCalledWith(ROUTES.BENUTZER);
    });
  });

  describe('handle unknown keycloak error', () => {
    it('should show error message', fakeAsync(() => {
      service._handleUnknownKeycloakError(createKeycloakHttpErrorResponse());

      expect(snackBarService.showError).toHaveBeenCalledWith('Der Benutzer konnte nicht angelegt werden.');
    }));
  });

  describe('getRoles', () => {
    it('should return no roles when none are active', () => {
      const result: string[] = service._getRoles(UserFormService.ALFA_GROUP);

      expect(result).toEqual([]);
    });

    it('should return poststelle role when active', () => {
      alfaGroup.get(UserFormService.POSTSTELLE).setValue(true);

      const result: string[] = service._getRoles(UserFormService.ALFA_GROUP);

      expect(result).toEqual([UserFormService.POSTSTELLE]);
    });
  });

  describe('getActiveOrganisationsEinheiten', () => {
    it('should return no groups when none are active', () => {
      const result: string[] = service._getActiveOrganisationsEinheiten();

      expect(result).toEqual([]);
    });

    it('should return bauamt group when active', () => {
      const organisationsEinheit: AdminOrganisationsEinheit = createAdminOrganisationsEinheit();
      organisationsEinheitenGroup.addControl(organisationsEinheit.name, new FormControl(true));

      const result: string[] = service._getActiveOrganisationsEinheiten();

      expect(result).toEqual([organisationsEinheit.name]);
    });
  });

  describe('ngOnDestroy', () => {
    it('should unsubscribe from initOrganisationsEinheiten$', () => {
      service._initOrganisationsEinheiten$.unsubscribe = jest.fn();

      service.ngOnDestroy();

      expect(service._initOrganisationsEinheiten$.unsubscribe).toHaveBeenCalled();
    });

    it('should unsubscribe from initOrganisationsEinheiten$', () => {
      service._alfaGroupChanges.unsubscribe = jest.fn();

      service.ngOnDestroy();

      expect(service._alfaGroupChanges.unsubscribe).toHaveBeenCalled();
    });
  });

  describe('get userName', () => {
    it('should return form control value of userName', () => {
      service.form = new FormGroup({ [UserFormService.USERNAME]: new FormControl('userNameDummy') });

      const userName: string = service.getUserName();

      expect(userName).toBe('userNameDummy');
    });
  });

  describe('map keycloak error message to control name', () => {
    it('should map username is missing error message to username', () => {
      expect(service._mapKeycloakErrorMessageToControlName(KeycloakErrorMessage.USERNAME_IS_MISSING)).toEqual(
        UserFormService.USERNAME,
      );
    });

    it('should user name already exists error message to username', () => {
      expect(service._mapKeycloakErrorMessageToControlName(KeycloakErrorMessage.USERNAME_EXISTS)).toEqual(
        UserFormService.USERNAME,
      );
    });

    it('should map email exists error message to email', () => {
      expect(service._mapKeycloakErrorMessageToControlName(KeycloakErrorMessage.EMAIL_EXISTS)).toEqual(UserFormService.EMAIL);
    });

    it('should throw error', () => {
      expect(() => service._mapKeycloakErrorMessageToControlName('unknown' as KeycloakErrorMessage)).toThrowError();
    });
  });

  describe('map keycloak field name to control name', () => {
    it('should map username to username', () => {
      expect(service._mapKeycloakFieldNameToControlName(KeycloakFieldName.USERNAME)).toEqual(UserFormService.USERNAME);
    });

    it('should map email to email', () => {
      expect(service._mapKeycloakFieldNameToControlName(KeycloakFieldName.EMAIL)).toEqual(UserFormService.EMAIL);
    });

    it('should throw error', () => {
      expect(() => service._mapKeycloakFieldNameToControlName('unknown' as KeycloakFieldName)).toThrowError();
    });
  });
});
