/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
import { BinaryFileService } from '@alfa-client/binary-file-shared';
import { CommandOrder, CommandResource, CommandService } from '@alfa-client/command-shared';
import { NavigationService } from '@alfa-client/navigation-shared';
import { createEmptyStateResource, createStateResource, decodeUrlFromEmbedding, StateResource } from '@alfa-client/tech-shared';
import { Mock, mock, useFromMock } from '@alfa-client/test-utils';
import { SnackBarService } from '@alfa-client/ui';
import { VorgangResource, VorgangService, VorgangWithEingangResource } from '@alfa-client/vorgang-shared';
import { expect } from '@jest/globals';
import { getUrl } from '@ngxp/rest';
import { cold, hot } from 'jest-marbles';
import { CommandLinkRel } from 'libs/command-shared/src/lib/command.linkrel';
import { createCommandResource } from 'libs/command-shared/test/command';
import { createVorgangResource, createVorgangWithEingangResource } from 'libs/vorgang-shared/test/vorgang';
import {
  createWiedervorlage,
  createWiedervorlageListResource,
  createWiedervorlageResource,
} from 'libs/wiedervorlage-shared/test/wiedervorlage';
import { of, Subscription } from 'rxjs';
import { WiedervorlageLinkRel, WiedervorlageListLinkRel } from './wiedervorlage.linkrel';
import { WiedervorlageMessages } from './wiedervorlage.message';
import {
  Wiedervorlage,
  WIEDERVORLAGE_UPLOADED_ATTACHMENTS,
  WiedervorlageListResource,
  WiedervorlageResource,
} from './wiedervorlage.model';
import { WiedervorlageRepository } from './wiedervorlage.repository';
import { WiedervorlageService } from './wiedervorlage.service';

import * as TechShared from '@alfa-client/tech-shared';

jest.mock('@alfa-client/tech-shared', () => mockAsEsModule('@alfa-client/tech-shared'));

function mockAsEsModule(module: string) {
  return {
    __esModule: true,
    ...jest.requireActual(module),
  };
}

describe('WiedervorlageService', () => {
  let service: WiedervorlageService;
  let repository: Mock<WiedervorlageRepository>;
  let navigationService: Mock<NavigationService>;
  let commandService: Mock<CommandService>;
  let snackbarService: Mock<SnackBarService>;
  let binaryFileService: Mock<BinaryFileService>;
  let vorgangService: Mock<VorgangService>;

  const vorgangResource: VorgangWithEingangResource = createVorgangWithEingangResource();
  const vorgangWithEingangStateResource: StateResource<VorgangWithEingangResource> = createStateResource(vorgangResource);

  const wiedervorlageResource: WiedervorlageResource = createWiedervorlageResource();
  const wiedervorlage: Wiedervorlage = createWiedervorlage();

  const commandResource: CommandResource = createCommandResource();
  const commandResourceWithEffectedResourceLink: CommandResource = createCommandResource([CommandLinkRel.EFFECTED_RESOURCE]);

  beforeEach(() => {
    repository = mock(WiedervorlageRepository);
    navigationService = mock(NavigationService);
    commandService = mock(CommandService);
    snackbarService = mock(SnackBarService);
    binaryFileService = mock(BinaryFileService);
    vorgangService = {
      ...mock(VorgangService),
      getVorgangWithEingang: jest.fn().mockReturnValue(of(vorgangWithEingangStateResource)),
    };

    navigationService.urlChanged.mockReturnValue(of({}));

    service = new WiedervorlageService(
      useFromMock(repository),
      useFromMock(navigationService),
      useFromMock(commandService),
      useFromMock(snackbarService),
      useFromMock(binaryFileService),
      useFromMock(vorgangService),
    );
  });

  describe('getWiedervorlageList', () => {
    const wiedervorlageList: WiedervorlageListResource = createWiedervorlageListResource();
    const wiedervorlageListStateResource: StateResource<WiedervorlageListResource> = createStateResource(wiedervorlageList);

    const vorgangWithEingangStateResource: StateResource<VorgangWithEingangResource> = createStateResource(vorgangResource);

    it('should call vorgangFacade getVorgangWithEingang', () => {
      service.getWiedervorlageList();

      expect(vorgangService.getVorgangWithEingang).toHaveBeenCalled();
    });

    it('should return values', () => {
      vorgangService.getVorgangWithEingang.mockReturnValue(hot('-a', { a: vorgangWithEingangStateResource }));
      service.wiedervorlageList$.asObservable = jest.fn().mockReturnValue(hot('-a', { a: wiedervorlageListStateResource }));

      const result = service.getWiedervorlageList();

      expect(result).toBeObservable(cold('ab', { a: createEmptyStateResource(true), b: wiedervorlageListStateResource }));
    });

    it.skip('should call loadWiedervorlagenByVorgang if required', (done) => {
      service._loadWiedervorlagenByVorgang = jest.fn();
      // service.wiedervorlageList$.next({ ...wiedervorlageListStateResource, reload: true });

      service.getWiedervorlageList().subscribe((list) => {
        expect(service._loadWiedervorlagenByVorgang).toHaveBeenCalled();
        done();
      });
      // expect(service.loadWiedervorlagenByVorgang).toHaveBeenCalled();
      service.wiedervorlageList$.next({ ...wiedervorlageListStateResource, reload: true });
      // service.wiedervorlageList$.next({ ...wiedervorlageListStateResource, reload: true });
      service.getWiedervorlageList();
    });

    it('should NOT call loadWiedervorlagenByVorgang if already loaded', () => {
      service._loadWiedervorlagenByVorgang = jest.fn();
      service.wiedervorlageList$.asObservable = jest.fn().mockReturnValue(of(wiedervorlageListStateResource));

      service.getWiedervorlageList();

      expect(service._loadWiedervorlagenByVorgang).not.toHaveBeenCalled();
    });
  });

  describe('getWiedervorlageListByGivenVorgang', () => {
    const vorgang: VorgangResource = createVorgangResource();
    const wiedervorlageListResource: WiedervorlageListResource = createWiedervorlageListResource();
    const wiedervorlageListStateResource: StateResource<WiedervorlageListResource> =
      createStateResource(wiedervorlageListResource);

    beforeEach(() => {
      service._refreshVorgangSubscription = jest.fn();
    });

    it('should call refresh vorgang subscription', () => {
      service.getWiedervorlageListByGivenVorgang(vorgang);

      expect(service._refreshVorgangSubscription).toHaveBeenCalled();
    });

    it('should return initial value', () => {
      service.wiedervorlageList$.asObservable = jest.fn().mockReturnValue(hot('-a', { a: wiedervorlageListResource }));

      const result = service.getWiedervorlageListByGivenVorgang(vorgang);

      expect(result).toBeObservable(cold('-a', { a: wiedervorlageListResource }));
    });

    it('should load wiedervorlagen by vorgang if required', () => {
      service._loadWiedervorlagenByVorgang = jest.fn();
      service.wiedervorlageList$.next(createEmptyStateResource());

      service.getWiedervorlageListByGivenVorgang(vorgang).subscribe();

      expect(service._loadWiedervorlagenByVorgang).toHaveBeenCalledWith(vorgang);
    });

    it('should NOT load wiedervorlagen by vorgang if already loaded', () => {
      service._loadWiedervorlagenByVorgang = jest.fn();
      service.wiedervorlageList$.next(wiedervorlageListStateResource);

      service.getWiedervorlageListByGivenVorgang(vorgang).subscribe();

      expect(service._loadWiedervorlagenByVorgang).not.toHaveBeenCalled();
    });
  });

  describe('refresh vorgang subscription', () => {
    beforeEach(() => {
      service.vorgangSubscription = <any>mock(Subscription);
      service._listenToVorgangChange = jest.fn();
    });

    it('should unsubscribe existing subscription', () => {
      service._refreshVorgangSubscription();

      expect(service.vorgangSubscription.unsubscribe).toHaveBeenCalled();
    });
    it('should call listen to vorgang change', () => {
      service._refreshVorgangSubscription();

      expect(service._listenToVorgangChange).toHaveBeenCalled();
    });
  });

  describe('getWiedervorlage', () => {
    const wiedervorlage: WiedervorlageResource = createWiedervorlageResource();
    const wiedervorlageStateResource: StateResource<WiedervorlageResource> = createStateResource(wiedervorlage);

    beforeEach(() => {
      navigationService.getDecodedParam.mockReturnValue(getUrl(wiedervorlageResource));
      repository.getWiedervorlage.mockReturnValue(of(wiedervorlageResource));
    });

    it('should return initial values', () => {
      service.wiedervorlage$.asObservable = jest.fn().mockReturnValue(hot('-a', { a: wiedervorlageStateResource }));

      const result = service.getWiedervorlage();

      expect(result).toBeObservable(cold('ab', { a: createEmptyStateResource(true), b: wiedervorlageStateResource }));
    });

    it('should set loading to true', () => {
      service._setWiedervorlageLoading = jest.fn();

      service.getWiedervorlage().subscribe();

      expect(service._setWiedervorlageLoading).toHaveBeenCalled();
    });

    it('should call navigationService', () => {
      service.getWiedervorlage().subscribe();

      expect(navigationService.getDecodedParam).toHaveBeenCalledWith(WiedervorlageService.encodedWiedervorlageUriParam);
    });

    it('should call repository', () => {
      service.getWiedervorlage().subscribe();

      expect(repository.getWiedervorlage).toHaveBeenCalledWith(getUrl(wiedervorlageResource));
    });

    it('should set returnvalue', () => {
      service.getWiedervorlage().subscribe();

      expect(service.wiedervorlage$.value).toEqual(createStateResource(wiedervorlageResource));
    });
  });

  describe('saveWiedervorlage', () => {
    beforeEach(() => {
      commandService.createCommand.mockReturnValue(of(createStateResource(commandResourceWithEffectedResourceLink)));
    });

    it('should return initial value and mapped value', () => {
      commandService.createCommand.mockReturnValue(
        hot('-a', { a: createStateResource(commandResourceWithEffectedResourceLink) }),
      );

      const result = service.saveWiedervorlage(wiedervorlageResource, wiedervorlage);

      expect(result).toBeObservable(
        cold('ab', {
          a: createEmptyStateResource(true),
          b: createStateResource(commandResourceWithEffectedResourceLink),
        }),
      );
    });

    it('should call command service', () => {
      service.saveWiedervorlage(wiedervorlageResource, wiedervorlage).subscribe();

      expect(commandService.createCommand).toHaveBeenCalledWith(wiedervorlageResource, WiedervorlageLinkRel.EDIT, {
        order: CommandOrder.EDIT_WIEDERVORLAGE,
        body: wiedervorlage,
      });
    });

    it('should set submit-in-progress on loading', () => {
      (<any>service).setSubmitInProgressLoading = jest.fn();

      service.saveWiedervorlage(wiedervorlageResource, wiedervorlage);

      expect((<any>service).setSubmitInProgressLoading).toHaveBeenCalled();
    });

    it('should call proceedAfterRecieveCommand', () => {
      service._proceedAfterReceiveCommand = jest.fn();

      service.saveWiedervorlage(wiedervorlageResource, wiedervorlage).subscribe();

      const expectedMessage = WiedervorlageMessages.SAVED.replace('{betreff}', wiedervorlage.betreff);
      expect(service._proceedAfterReceiveCommand).toHaveBeenCalledWith(
        createStateResource(commandResourceWithEffectedResourceLink),
        expectedMessage,
      );
    });
  });

  describe('createWiedervorlage', () => {
    beforeEach(() => {
      commandService.createCommand.mockReturnValue(of(createStateResource(commandResource)));
    });

    it('should return intitial value and mapped value', () => {
      commandService.createCommand.mockReturnValue(hot('-a', { a: createStateResource(commandResource) }));

      const result = service.createWiedervorlage(wiedervorlage);

      expect(result).toBeObservable(cold('ab', { a: createEmptyStateResource(true), b: createStateResource(commandResource) }));
    });

    it('should set submit-in-progress on loading', () => {
      (<any>service).setSubmitInProgressLoading = jest.fn();

      service.createWiedervorlage(wiedervorlage);

      expect((<any>service).setSubmitInProgressLoading).toHaveBeenCalled();
    });

    it('should call command service', () => {
      service.createWiedervorlage(wiedervorlage).subscribe();

      const command = { order: CommandOrder.CREATE_WIEDERVORLAGE, body: wiedervorlage };
      expect(commandService.createCommand).toHaveBeenCalledWith(
        service.wiedervorlageList$.value.resource,
        WiedervorlageListLinkRel.CREATE_WIEDERVORLAGE,
        command,
      );
    });

    it('should call proceedAfterReceiveCommand', () => {
      service._proceedAfterReceiveCommand = jest.fn();

      service.createWiedervorlage(wiedervorlage).subscribe();

      const expectedMessage = WiedervorlageMessages.CREATED.replace('{betreff}', wiedervorlage.betreff);
      expect(service._proceedAfterReceiveCommand).toHaveBeenCalledWith(createStateResource(commandResource), expectedMessage);
    });
  });

  describe('proceedAfterRecieveCommand', () => {
    beforeEach(() => {
      service.clearUploadedFiles = jest.fn();
    });

    it('should set reload on wiedervorlageList', () => {
      service._proceedAfterReceiveCommand(createStateResource(commandResourceWithEffectedResourceLink), null);

      expect(service.wiedervorlageList$.value.reload).toBe(true);
    });

    it('should do nothing on apiError', () => {
      service._proceedAfterReceiveCommand(createEmptyStateResource(), null);

      expect(service.wiedervorlageList$.value.reload).toBe(false);
    });

    it('should call show snackbar', () => {
      snackbarService.show = jest.fn();
      const snackbarMessage: string = 'XX wurde XX';

      service._proceedAfterReceiveCommand(createStateResource(commandResourceWithEffectedResourceLink), snackbarMessage);

      expect(snackbarService.show).toHaveBeenCalledWith(commandResourceWithEffectedResourceLink, snackbarMessage);
    });

    it('should clear uploaded files', () => {
      service._proceedAfterReceiveCommand(createStateResource(commandResourceWithEffectedResourceLink), null);

      expect(service.clearUploadedFiles).toHaveBeenCalled();
    });
  });

  describe('order actions', () => {
    const wiedervorlageResource: WiedervorlageResource = createWiedervorlageResource();

    beforeEach(() => {
      commandService.createCommand.mockReturnValue(of(createStateResource(commandResourceWithEffectedResourceLink)));
      (<any>service).reloadWiedervorlageList = jest.fn();
      (<any>service).loadAndSetWiedervorlageByUrl = jest.fn();

      service.wiedervorlage$.next(createStateResource(wiedervorlageResource));
      service.clearUploadedFiles = jest.fn();
    });

    describe('wiedervorlage erledigen', () => {
      it('should call commandService', () => {
        service.erledigen();

        expect(commandService.createCommand).toHaveBeenCalledWith(wiedervorlageResource, WiedervorlageLinkRel.ERLEDIGEN, {
          order: CommandOrder.WIEDERVORLAGE_ERLEDIGEN,
          body: null,
        });
      });

      it('should reload wiedervorlage', () => {
        service.erledigen();

        expect((<any>service).loadAndSetWiedervorlageByUrl).toHaveBeenCalled();
      });

      it('should show snackbar', () => {
        service.erledigen();

        expect(snackbarService.show).toHaveBeenCalledWith(
          commandResourceWithEffectedResourceLink,
          `Die Wiedervorlage ${wiedervorlageResource.betreff} wurde erledigt`,
        );
      });

      it('should clear uploaded files', () => {
        service.erledigen();

        expect(service.clearUploadedFiles).toHaveBeenCalled();
      });
    });

    describe('wiedervorlage wiedereroeffnen', () => {
      it('should call commandService', () => {
        service.wiedereroeffnen();

        expect(commandService.createCommand).toHaveBeenCalledWith(wiedervorlageResource, WiedervorlageLinkRel.WIEDEREROEFFNEN, {
          order: CommandOrder.WIEDERVORLAGE_WIEDEREROEFFNEN,
          body: null,
        });
      });

      it('should reload wiedervorlage', () => {
        service.wiedereroeffnen();

        expect((<any>service).loadAndSetWiedervorlageByUrl).toHaveBeenCalled();
      });

      it('should show snackbar', () => {
        service.wiedereroeffnen();

        expect(snackbarService.show).toHaveBeenCalledWith(
          commandResourceWithEffectedResourceLink,
          `Die Wiedervorlage ${wiedervorlageResource.betreff} wurde wiedereröffnet`,
        );
      });

      it('should clear uploaded files', () => {
        service.wiedereroeffnen();

        expect(service.clearUploadedFiles).toHaveBeenCalled();
      });
    });
  });

  describe('showSnackbar', () => {
    const message: string = 'SnackbarMessage';

    it('should call service if command is done', () => {
      const commandResource: CommandResource = createCommandResource([CommandLinkRel.EFFECTED_RESOURCE]);

      service._showSnackBar(commandResource, message);

      expect(snackbarService.show).toHaveBeenCalledWith(commandResource, message);
    });

    it('should NOT call service if command is pending', () => {
      const commandResource: CommandResource = createCommandResource();

      service._showSnackBar(commandResource, message);

      expect(snackbarService.show).not.toHaveBeenCalled();
    });
  });

  describe('on navigation', () => {
    beforeEach(() => {
      (<any>service).navigateToVorgangDetailPage = jest.fn();

      service._setWiedervorlageListReload = jest.fn();
      (<any>service).forceWiedervorlageReload = jest.fn();
      service._clearWiedervorlagenList = jest.fn();
      service.clearUploadedFiles = jest.fn();
    });

    it('should clear uploaded files', () => {
      service._onNavigation({});

      expect(service.clearUploadedFiles).toHaveBeenCalled();
    });

    describe('to vorgang detail', () => {
      beforeEach(() => {
        (<any>service).navigateToVorgangDetailPage.mockReturnValue(true);
      });

      it('should set wiedervorlage list on reload', () => {
        service._onNavigation({});

        expect(service._setWiedervorlageListReload).toHaveBeenCalled();
      });

      it('should set wiedervorlage on reload', () => {
        service._onNavigation({});

        expect((<any>service).forceWiedervorlageReload).toHaveBeenCalled();
      });
    });

    describe('to vorgang list', () => {
      it('should clear list', () => {
        service._onNavigation({});

        expect(service._clearWiedervorlagenList).toHaveBeenCalled();
      });
    });
  });

  describe('loadWiedervorlagenByVorgangStrict', () => {
    const wiedervorlageListResource: WiedervorlageListResource = createWiedervorlageListResource();

    beforeEach(() => {
      repository.getWiedervorlageList.mockReturnValue(hot('-a', { a: wiedervorlageListResource }));
    });

    it('should return value', () => {
      const result = service.loadWiedervorlagenByVorgangStrict(vorgangResource);

      expect(result).toBeObservable(
        cold('ab', {
          a: createEmptyStateResource(true),
          b: createStateResource(wiedervorlageListResource),
        }),
      );
    });

    it('should call repository', () => {
      service.loadWiedervorlagenByVorgangStrict(vorgangResource);

      expect(repository.getWiedervorlageList).toHaveBeenCalledWith(vorgangResource);
    });
  });

  describe('hasEditLink', () => {
    beforeEach(() => {
      navigationService.getDecodedParam.mockReturnValue(getUrl(wiedervorlageResource));
      repository.getWiedervorlage.mockReturnValue(of(wiedervorlageResource));
    });

    it('should call decodeUrlFromEmbedding', () => {
      const spy = jest.spyOn(TechShared, 'decodeUrlFromEmbedding');

      service.hasEditLink(getUrl(wiedervorlageResource));

      expect(spy).toHaveBeenCalled();
    });

    it('should call repository.getWiedervorlage', () => {
      const url: string = getUrl(wiedervorlageResource);
      const decodedUrl: string = decodeUrlFromEmbedding(url);

      service.hasEditLink(url);

      expect(repository.getWiedervorlage).toHaveBeenCalledWith(decodedUrl);
    });
  });

  describe('clearUploadedFiles', () => {
    it('should call binary file service', () => {
      service.clearUploadedFiles();

      expect(binaryFileService.clearUploadedFiles).toHaveBeenCalledWith(WIEDERVORLAGE_UPLOADED_ATTACHMENTS);
    });
  });

  describe('listen to vorgang change', () => {
    beforeEach(() => {
      service._handleVorgangChange = jest.fn();
    });

    it('should call vorgang service to get vorgang with eingang', () => {
      service._listenToVorgangChange();

      expect(vorgangService.selectVorgangWithEingang).toHaveBeenCalled();
    });

    it('should call handle vorgang change', () => {
      service._listenToVorgangChange();

      expect(service._handleVorgangChange).toHaveBeenCalledWith(vorgangWithEingangStateResource);
    });
  });

  describe('handle vorgang change', () => {
    it('should set reload flag if list is loaded and state resource is loading', () => {
      service.shouldReload = false;
      service.wiedervorlageList$.next(createStateResource(createWiedervorlageListResource()));

      service._handleVorgangChange({ ...vorgangWithEingangStateResource, loading: true });

      expect(service.shouldReload).toBeTruthy();
    });

    it('should set reload flag if list is loaded and state resource is reloading', () => {
      service.wiedervorlageList$.next(createStateResource(createWiedervorlageListResource()));
      service.shouldReload = false;

      service._handleVorgangChange({ ...vorgangWithEingangStateResource, reload: true });

      expect(service.shouldReload).toBeTruthy();
    });

    describe('on setted reload flag', () => {
      beforeEach(() => {
        service._setWiedervorlageListReload = jest.fn();
        service.shouldReload = true;
      });

      describe('and loaded vorgang resource', () => {
        it('should call set wiedervorlage list reload', () => {
          service._handleVorgangChange(vorgangWithEingangStateResource);

          expect(service._setWiedervorlageListReload).toHaveBeenCalled();
        });

        it('and loaded vorgang resource should call set wiedervorlage list reload', () => {
          service._handleVorgangChange(vorgangWithEingangStateResource);

          expect(service.shouldReload).toBeFalsy();
        });
      });
    });
  });
});
