/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
import { HttpErrorResponse, HttpStatusCode } from '@angular/common/http';
import { TestBed } from '@angular/core/testing';
import {
  ActivatedRouteSnapshot,
  CanActivateFn,
  RouterStateSnapshot,
  UrlSegment,
} from '@angular/router';
import { faker } from '@faker-js/faker';
import { encodeUrlForEmbedding } from '@alfa-client/tech-shared';
import { mock } from '@alfa-client/test-utils';
import { SnackBarService } from '@alfa-client/ui';
import { WiedervorlageService } from '@alfa-client/wiedervorlage-shared';
import { EMPTY, Observable, of } from 'rxjs';

import * as RouterHelper from '@angular/router';
import * as Guard from './wiedervorlage.guard';

const next: ActivatedRouteSnapshot = {
  params: { wiedervorlageUrl: '', vorgangWithEingangUrl: '' },
} as unknown as ActivatedRouteSnapshot;
const state: RouterStateSnapshot = {
  root: { url: [<UrlSegment>{}] },
} as unknown as RouterStateSnapshot;

describe('wiedervorlageGuard', () => {
  const wiedervorlageService = mock(WiedervorlageService);
  const snackbarService = mock(SnackBarService);
  const createUrlTreeForVorgang = jest
    .spyOn(Guard, 'createUrlTreeForVorgang')
    .mockReturnValue(<RouterHelper.UrlTree>{});

  beforeEach(() => {
    TestBed.configureTestingModule({
      providers: [
        {
          provide: WiedervorlageService,
          useValue: wiedervorlageService,
        },
        {
          provide: SnackBarService,
          useValue: snackbarService,
        },
      ],
    });

    jest.spyOn(wiedervorlageService, 'hasEditLink').mockReturnValue(EMPTY);
  });

  it('should be created', () => {
    const guard = TestBed.runInInjectionContext(() =>
      Guard.wiedervorlageGuard(next, state),
    ) as unknown as CanActivateFn;

    expect(guard).toBeTruthy();
  });

  describe('wiedervorlageGuard', () => {
    it('should call wiedervorlageService.hasEditLink', () => {
      const spy = jest.spyOn(wiedervorlageService, 'hasEditLink');
      next.params['wiedervorlageUrl'] = encodeUrlForEmbedding(faker.internet.url());

      TestBed.runInInjectionContext(() =>
        Guard.wiedervorlageGuard(next, state),
      ) as unknown as CanActivateFn;

      expect(spy).toHaveBeenCalled();
    });

    it('should return UrlTree object', (done) => {
      jest.spyOn(wiedervorlageService, 'hasEditLink').mockReturnValue(of(false));

      TestBed.runInInjectionContext(() =>
        (Guard.wiedervorlageGuard(next, state) as Observable<RouterHelper.UrlTree>).subscribe(
          (result) => {
            expect(result).toBeInstanceOf(Object);
            done();
          },
        ),
      ) as unknown as CanActivateFn;
    });

    it('should return true', (done) => {
      jest.spyOn(wiedervorlageService, 'hasEditLink').mockReturnValue(of(true));

      TestBed.runInInjectionContext(() =>
        (Guard.wiedervorlageGuard(next, state) as Observable<RouterHelper.UrlTree>).subscribe(
          (result) => {
            expect(result).toBeTruthy();
            done();
          },
        ),
      ) as unknown as CanActivateFn;
    });

    it.skip('should call handleError if error occurs', (done) => {
      jest.spyOn(wiedervorlageService, 'hasEditLink').mockReturnValue(of(new Error()));
      const spy = jest.spyOn(Guard, 'handleError');

      TestBed.runInInjectionContext(() =>
        (Guard.wiedervorlageGuard(next, state) as Observable<RouterHelper.UrlTree>).subscribe(
          () => {
            expect(spy).toHaveBeenCalled();
            done();
          },
        ),
      ) as unknown as CanActivateFn;
    });
  });

  describe('handleLinkCheck', () => {
    it('should return true if hasLink is true', () => {
      const hasLink = true;

      const result = Guard.handleLinkCheck(hasLink, null, null);

      expect(result).toBeTruthy();
    });

    it('should call showSnackbar if hasLink is false', () => {
      const hasLink = false;
      const spy = jest.spyOn(Guard, 'showSnackbar');

      Guard.handleLinkCheck(hasLink, next, snackbarService as unknown as SnackBarService);

      expect(spy).toHaveBeenCalled();
    });

    it('should call createUrlTreeForVorgang if hasLink is false', () => {
      const hasLink = false;
      jest.spyOn(Guard, 'showSnackbar');

      Guard.handleLinkCheck(hasLink, next, snackbarService as unknown as SnackBarService);

      expect(createUrlTreeForVorgang).toHaveBeenCalled();
    });
  });

  describe('showSnackbar', () => {
    const showError = jest.spyOn(snackbarService, 'showError');

    it('should call snackbarService.showError', () => {
      Guard.showSnackbar(snackbarService as unknown as SnackBarService);

      expect(showError).toHaveBeenCalledWith(
        'Im Status "Zu löschen" ist die Bearbeitung von Wiedervorlagen nicht möglich.',
      );
    });
  });

  describe('handleError', () => {
    it('should return Observable<UrlTree> if HTTP Status 403', () => {
      const error = createHttpError(HttpStatusCode.Forbidden);

      const res: RouterHelper.UrlTree = Guard.handleError(error, next);

      expect(res).toBeInstanceOf(Object);
    });

    it('should return nothing if HTTP Status is not 403', () => {
      const error = createHttpError(HttpStatusCode.Unauthorized);

      function monitorErrorThrowing() {
        Guard.handleError(error, next);
      }

      expect(monitorErrorThrowing).toThrow();
    });
  });

  function createHttpError(status: HttpStatusCode): any {
    return { error: new HttpErrorResponse({ status }) };
  }
});
